"""
Parse a junit report file into a family of objects
"""
from __future__ import unicode_literals

from typing import TYPE_CHECKING

import os
import sys
import xml.etree.ElementTree as ET
import collections
import uuid

from .case_result import CaseResult
from .render import HTMLReport, JunitLoaderBase
from .textutils import unicode_str

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Union, Any, OrderedDict

NO_CLASSNAME = "no-testclass"
PASSED = CaseResult.PASSED
FAILED = CaseResult.FAILED
SKIPPED = CaseResult.SKIPPED
ABSENT = CaseResult.ABSENT
UNKNOWN = CaseResult.UNKNOWN


def clean_xml_attribute(element: "ET.Element", attribute: str, default: "Optional[str]"=None):
    """
    Get an XML attribute value and ensure it is legal in XML
    :param element:
    :param attribute:
    :param default:
    :return:
    """

    value = element.attrib.get(attribute, default)
    if value:
        value = value.encode("utf-8", errors="replace").decode("utf-8", errors="backslashreplace")
        value = value.replace(u"\ufffd", "?")  # strip out the unicode replacement char

    return value


class ParserError(Exception):
    """
    We had a problem parsing a file
    """
    def __init__(self, message: str):
        super(ParserError, self).__init__(message)


class ToJunitXmlBase(object):
    """
    Base class of all objects that can be serialized to Junit XML
    """
    def tojunit(self) -> "ET.Element":
        """
        Return an Element matching this object
        :return:
        """
        raise NotImplementedError()

    def make_element(self, xmltag: str, text: "Optional[str]"=None, attribs: "Optional[Dict[str, Any]]"=None):
        """
        Create an Element and put text and/or attribs into it
        :param xmltag: tag name
        :param text:
        :param attribs: dict of xml attributes
        :return:
        """
        element = ET.Element(unicode_str(xmltag))
        if text is not None:
            element.text = unicode_str(text)
        if attribs is not None:
            for item in attribs:
                element.set(unicode_str(item), unicode_str(attribs[item]))
        return element


class AnchorBase(object):
    """
    Base class that can generate a unique anchor name.
    """
    def __init__(self):
        self._anchor = None

    def id(self):
        return self.anchor()

    def anchor(self):
        """
        Generate a html anchor name
        :return:
        """
        if not self._anchor:
            self._anchor = str(uuid.uuid4())
        return self._anchor


class Class(AnchorBase):
    """
    A namespace for a test
    """
    name: "Optional[str]" = None
    cases: "list[Case]"
    
    def __init__(self):
        super(Class, self).__init__()
        self.cases = list()


class Property(AnchorBase, ToJunitXmlBase):
    """
    Test Properties
    """
    def __init__(self):
        super(Property, self).__init__()
        self.name: "Optional[str]" = None
        self.value: "Optional[str]" = None

    def tojunit(self):
        """
        Return the xml element for this property
        :return:
        """
        prop = self.make_element("property")
        prop.set(u"name", unicode_str(self.name))
        prop.set(u"value", unicode_str(self.value))
        return prop


class Case(AnchorBase, ToJunitXmlBase):
    """
    Test cases
    """
    failure: "Optional[str]" = None
    failure_msg: "Optional[str]" = None
    skipped: "Optional[str]" = None
    skipped_msg: "Optional[str]" = None
    stderr: "Optional[Union[str,Any]]" = None
    stdout: "Optional[Union[str,Any]]" = None
    duration: float = 0
    name: "Optional[str]" = None
    testclass: "Optional[Class]" = None
    properties: "List[Property]"

    def __init__(self):
        super(Case, self).__init__()
        self.properties = list()

    @property
    def display_suffix(self):
        if self.skipped:
            return "[s]"
        return ""

    def outcome(self) -> CaseResult:
        """
        Return the result of this test case
        :return:
        """
        if self.skipped:
            return CaseResult.SKIPPED
        elif self.failed():
            return CaseResult.FAILED
        return CaseResult.PASSED

    def prefix(self):
        if self.skipped:
            return "[S]"
        if self.failed():
            return "[F]"
        return ""

    def tojunit(self):
        """
        Turn this test case back into junit xml
        :note: this may not be the exact input we loaded
        :return:
        """
        if self.testclass is None or self.testclass.name is None:
            testclass_name = ""
        else:
            testclass_name = self.testclass.name

        testcase = self.make_element("testcase")
        testcase.set(u"name", unicode_str(self.name))
        testcase.set(u"classname", unicode_str(testclass_name))
        testcase.set(u"time", unicode_str(self.duration))

        if self.stderr is not None:
            testcase.append(self.make_element("system-err", self.stderr))
        if self.stdout is not None:
            testcase.append(self.make_element("system-out", self.stdout))

        if self.failure is not None:
            testcase.append(self.make_element(
                "failure", self.failure,
                {
                    "message": self.failure_msg
                }))

        if self.skipped:
            testcase.append(self.make_element(
                "skipped", self.skipped,
                {
                    "message": self.skipped_msg
                }))

        if self.properties:
            props = self.make_element("properties")
            for prop in self.properties:
                props.append(prop.tojunit())
            testcase.append(props)

        return testcase

    def fullname(self):
        """
        Get the full name of a test case
        :return:
        """
        if self.testclass is None or self.testclass.name is None:
            testclass_name = ""
        else:
            testclass_name = self.testclass.name
        return "{} : {}".format(testclass_name, self.name)

    def basename(self):
        """
        Get a short name for this case
        :return:
        """
        if (   self.name is None
            or self.testclass is None
            or self.testclass.name is None
        ):
            return None

        if self.name.startswith(self.testclass.name):
            return self.name[len(self.testclass.name):]
        return self.name

    def failed(self):
        """
        Return True if this test failed
        :return:
        """
        return self.failure is not None


class Suite(AnchorBase, ToJunitXmlBase):
    """
    Contains test cases (usually only one suite per report)
    """
    name: "Optional[str]" = None
    properties: "List[Property]"
    classes: "OrderedDict[str, Class]"
    duration: float = 0
    package: "Optional[str]" = None
    errors: "List[Dict[str, Optional[Union[str,Any]]]]"
    stdout: "Optional[Union[str,Any]]" = None
    stderr: "Optional[Union[str,Any]]" = None

    def __init__(self):
        super(Suite, self).__init__()
        self.classes = collections.OrderedDict()
        self.properties = []
        self.errors = []

    def tojunit(self):
        """
        Return an element for this whole suite and all it's cases
        :return:
        """
        suite = self.make_element("testsuite")
        suite.set(u"name", unicode_str(self.name))
        suite.set(u"time", unicode_str(self.duration))
        if self.properties:
            props = self.make_element("properties")
            for prop in self.properties:
                props.append(prop.tojunit())
            suite.append(props)

        for testcase in self.all():
            suite.append(testcase.tojunit())
        return suite

    def __contains__(self, item: str):
        """
        Return True if the given test classname is part of this test suite
        :param item:
        :return:
        """
        return item in self.classes

    def __getitem__(self, item: str):
        """
        Return the given test class object
        :param item:
        :return:
        """
        return self.classes[item]

    def __setitem__(self, key: str, value: "Class"):
        """
        Add a test class
        :param key:
        :param value:
        :return:
        """
        self.classes[key] = value

    def all(self):
        """
        Return all testcases
        :return:
        """
        tests: "List[Case]" = list()
        for testclass in self.classes:
            tests.extend(self.classes[testclass].cases)
        return tests

    def failed(self):
        """
        Return all the failed testcases
        :return:
        """
        return [test for test in self.all() if test.failed()]

    def skipped(self):
        """
        Return all skipped testcases
        :return:
        """
        return [test for test in self.all() if test.skipped]

    def passed(self):
        """
        Return all the passing testcases
        :return:
        """
        return [test for test in self.all() if not test.failed() and not test.skipped]


class Junit(JunitLoaderBase):
    """
    Parse a single junit xml report
    """
    def __init__(self, filename: "Optional[str]"=None, xmlstring: "Optional[str]"=None):
        """
        Parse the file
        :param filename:
        :return:
        """
        super().__init__()
        self.suites: "List[Suite]" = []
        self.tree: "Optional[Union[ET.ElementTree,ET.Element]]" = None
        self.filename: "Optional[str]" = filename
        if filename == "-":
            # read the xml from stdin
            stdin = sys.stdin.read()
            xmlstring = stdin
            self.filename = None

        if self.filename is not None:
            self.tree = ET.parse(self.filename)
        elif xmlstring is not None:
            self._read(xmlstring)
        else:
            raise ValueError("Missing any filename or xmlstring")
        self.process()


    def __iter__(self):
        return self.suites.__iter__()

    def _read(self, xmlstring: str):
        """
        Populate the junit xml document tree from a string
        :param xmlstring:
        :return:
        """
        self.tree = ET.fromstring(xmlstring)


    def process(self):
        """
        populate the report from the xml
        :return:
        """
        testrun = False
        suites: "Optional[list[ET.Element]]" = None
        root: "ET.Element"
        if isinstance(self.tree, ET.ElementTree):
            root = self.tree.getroot()
        else:
            root = self.tree

        if root.tag == "testrun":
            testrun = True
            root: "ET.Element" = root[0]

        if root.tag == "testsuite":
            suites = [root]

        if root.tag == "testsuites" or testrun:
            suites = [x for x in root]

        if suites is None:
            raise ParserError("could not find test suites in results xml")
        suitecount = 0
        for suite in suites:
            suitecount += 1
            cursuite = Suite()
            self.suites.append(cursuite)
            cursuite.name = clean_xml_attribute(suite, "name", default="suite-" + str(suitecount))
            cursuite.package = clean_xml_attribute(suite, "package")

            cursuite.duration = float(suite.attrib.get("time", '0').replace(',', '') or '0')

            for element in suite:
                if element.tag == "error":
                    # top level error?
                    errtag = {
                        "message": element.attrib.get("message", ""),
                        "type": element.attrib.get("type", ""),
                        "text": element.text
                    }
                    cursuite.errors.append(errtag)
                if element.tag == "system-out":
                    cursuite.stdout = element.text
                if element.tag == "system-err":
                    cursuite.stderr = element.text

                if element.tag == "properties":
                    for prop in element:
                        if prop.tag == "property":
                            newproperty = Property()
                            newproperty.name = prop.attrib["name"]
                            newproperty.value = prop.attrib["value"]
                            cursuite.properties.append(newproperty)

                if element.tag == "testcase":
                    testcase = element

                    if not testcase.attrib.get("classname", None):
                        testcase.attrib["classname"] = NO_CLASSNAME

                    if testcase.attrib["classname"] not in cursuite:
                        testclass = Class()
                        testclass.name = testcase.attrib["classname"]
                        cursuite[testclass.name] = testclass

                    testclass: "Class" = cursuite[testcase.attrib["classname"]]
                    newcase = Case()
                    newcase.name = clean_xml_attribute(testcase, "name")
                    newcase.testclass = testclass
                    newcase.duration = float(testcase.attrib.get("time", '0').replace(',', '') or '0')
                    testclass.cases.append(newcase)

                    # does this test case have any children?
                    for child in testcase:
                        if child.tag == "skipped":
                            newcase.skipped = child.text
                            if "message" in child.attrib:
                                newcase.skipped_msg = child.attrib["message"]
                            if not newcase.skipped:
                               newcase.skipped = "skipped"
                        elif child.tag == "system-out":
                            newcase.stdout = child.text
                        elif child.tag == "system-err":
                            newcase.stderr = child.text
                        elif child.tag == "failure":
                            newcase.failure = child.text
                            if "message" in child.attrib:
                                newcase.failure_msg = child.attrib["message"]
                            if not newcase.failure:
                                newcase.failure = "failed"
                        elif child.tag == "error":
                            newcase.failure = child.text
                            if "message" in child.attrib:
                                newcase.failure_msg = child.attrib["message"]
                            if not newcase.failure:
                                newcase.failure = "error"
                        elif child.tag == "properties":
                            for prop in child:
                                newproperty = Property()
                                newproperty.name = prop.attrib["name"]
                                newproperty.value = prop.attrib["value"]
                                newcase.properties.append(newproperty)

    def html(self, show_toc: bool=True):
        """
        Render the test suite as a HTML report with links to errors first.
        :return:
        """

        doc = HTMLReport(show_toc=show_toc)
        title = "Test Results"
        if self.filename:
            if os.path.exists(self.filename):
                title = os.path.basename(self.filename)
        doc.load(self, title=title)
        return str(doc)



"""
Parse results files
"""
from .parser import Junit


def load_report(filename: str) -> Junit:
    """
    Load a report from disjk
    """
    return Junit(filename=filename)


def load_string(text: str) -> Junit:
    """
    Load a report from a string
    """
    return Junit(xmlstring=text)



"""
Handle multiple parsed junit reports
"""
from __future__ import unicode_literals

import os
from typing import TYPE_CHECKING

from jinja2 import Environment, select_autoescape

from . import parser
from .case_result import CaseResult
from .render import ReportContainer

UNTESTED = CaseResult.UNTESTED
PARTIAL_PASS = CaseResult.PARTIAL_PASS
PARTIAL_FAIL = CaseResult.PARTIAL_FAIL
TOTAL_FAIL = CaseResult.TOTAL_FAIL


if TYPE_CHECKING: # pragma: no cover
    from .parser import Case, Class
    from typing import Dict, List, Optional, Literal


class ReportMatrix(ReportContainer):
    """
    Load and handle several report files
    """
    cases: "Dict[str, Dict[str, Dict[str, Case]]]"
    classes: "Dict[str, Dict[str, Class]]"
    casenames: "Dict[str, List[str]]"
    result_stats: "Dict[CaseResult, int]"
    case_results: "Dict[str, Dict[str, List[CaseResult]]]"

    def __init__(self):
        super(ReportMatrix, self).__init__()
        self.cases = {}
        self.classes = {}
        self.casenames = {}
        self.result_stats = {}
        self.case_results = {}

    def add_case_result(self, case: "Case"):
        if case.testclass is None or case.testclass.name is None:
            testclass = ""
        else:
            testclass = case.testclass.name
        casename = "" if case.name is None else case.name
        if testclass not in self.case_results:
            self.case_results[testclass] = {}
        if casename not in self.case_results[testclass]:
            self.case_results[testclass][casename] = []
        self.case_results[testclass][casename].append(case.outcome())

    def report_order(self):
        return sorted(self.reports.keys())

    def short_outcome(self, outcome: CaseResult) -> "Literal['ok', '/', 's', 'f', 'F', '%', 'X', 'U', '?']":
        if outcome == CaseResult.PASSED:
            return "/"
        elif outcome == CaseResult.SKIPPED: # pragma: no cover
            return "s" # currently unused because SKIPPED returns UNTESTED
        elif outcome == CaseResult.FAILED:
            return "f"
        elif outcome == CaseResult.TOTAL_FAIL:
            return "F"
        elif outcome == CaseResult.PARTIAL_PASS:
            return "%"
        elif outcome == CaseResult.PARTIAL_FAIL:
            return "X"
        elif outcome == CaseResult.UNTESTED:
            return "U"

        return "?"

    def add_report(self, filename: str):
        """
        Load a report into the matrix
        :param filename:
        :return:
        """
        parsed = parser.Junit(filename=filename)
        filename = os.path.basename(filename)
        self.reports[filename] = parsed

        for suite in parsed.suites:
            for testclass in suite.classes:
                if testclass not in self.classes:
                    self.classes[testclass] = {}
                if testclass not in self.casenames:
                    self.casenames[testclass] = list()
                self.classes[testclass][filename] = suite.classes[testclass]

                for testcase in self.classes[testclass][filename].cases:
                    name = "" if testcase.name is None else testcase.name.strip()
                    if name not in self.casenames[testclass]:
                        self.casenames[testclass].append(name)

                    if testclass not in self.cases:
                        self.cases[testclass] = {}
                    if name not in self.cases[testclass]:
                        self.cases[testclass][name] = {}
                    self.cases[testclass][name][filename] = testcase

                    outcome = testcase.outcome()
                    self.add_case_result(testcase)

                    self.result_stats[outcome] = 1 + self.result_stats.get(
                        outcome, 0)

    def summary(self) -> str:
        """
        Render a summary of the matrix
        :return:
        """
        raise NotImplementedError()

    def combined_result_list(self, classname: str, casename: str):
        """
        Combone the result of all instances of the given case
        :param classname:
        :param casename:
        :return:
        """
        if classname in self.case_results:
            if casename in self.case_results[classname]:
                results = self.case_results[classname][casename]
                return self.combined_result(results)

        return " ", ""

    def combined_result(self, results: "List[CaseResult]"):
        """
        Given a list of results, produce a "combined" overall result
        :param results:
        :return:
        """
        if results:
            if CaseResult.PASSED in results:
                if CaseResult.FAILED in results:
                    return self.short_outcome(CaseResult.PARTIAL_FAIL), CaseResult.PARTIAL_FAIL.title()
                return self.short_outcome(CaseResult.PASSED), CaseResult.PASSED.title()

            if CaseResult.FAILED in results:
                return self.short_outcome(CaseResult.FAILED), CaseResult.FAILED.title()
            if CaseResult.SKIPPED in results:
                return self.short_outcome(CaseResult.UNTESTED), CaseResult.UNTESTED.title()
            if CaseResult.PARTIAL_PASS in results:
                return self.short_outcome(CaseResult.PARTIAL_PASS), CaseResult.PARTIAL_PASS.title()
            if CaseResult.TOTAL_FAIL in results:
                return self.short_outcome(CaseResult.TOTAL_FAIL), CaseResult.TOTAL_FAIL.title()
        return " ", ""


class HTMLMatrix:
    def __init__(self, matrix: "ReportMatrix"):
        self.matrix = matrix
        self.title: str = "JUnit Matrix"

    def __str__(self) -> str:
        loader = self.matrix.loader_factory.get_loader()
        env = Environment(
            loader=loader,
            autoescape=select_autoescape(["html"])
        )
        template = env.get_template("matrix.html")
        return template.render(matrix=self.matrix, title=self.title)


class HtmlReportMatrix(ReportMatrix):
    """
    Render a matrix report as html
    """

    outdir: str

    def __init__(self, outdir: str):
        super(HtmlReportMatrix, self).__init__()
        self.outdir = outdir

    def add_report(self, filename: str, show_toc: bool=True):
        """
        Load a report
        """
        super(HtmlReportMatrix, self).add_report(filename)
        basename = os.path.basename(filename)
        # make the individual report too
        report = self.reports[basename].html(show_toc=show_toc)
        if self.outdir != "" and not os.path.exists(self.outdir):
            os.makedirs(self.outdir)
        with open(
                os.path.join(self.outdir, basename) + ".html", "wb") as filehandle:
            filehandle.write(report.encode("utf-8"))

    def short_outcome(self, outcome: CaseResult) -> "Literal['ok', '/', 's', 'f', 'F', '%', 'X', 'U', '?']":
        if outcome == CaseResult.PASSED:
            return "ok"
        return super(HtmlReportMatrix, self).short_outcome(outcome)

    def short_axis(self, axis: str):
        if axis.endswith(".xml"):
            return axis[:-4]
        return axis

    def summary(self):
        """
        Render the html
        :return:
        """
        html_matrix = HTMLMatrix(self)

        return str(html_matrix)


class TextReportMatrix(ReportMatrix):
    """
    Render a matrix report as text
    """

    def summary(self):
        """
        Render as a string
        :return:
        """

        output = "\nMatrix Test Report\n"
        output += "===================\n"

        axis = list(self.reports.keys())
        axis.sort()

        # find the longest classname or test case name
        left_indent = 0
        for classname in self.classes:
            left_indent = max(len(classname), left_indent)
            for casename in self.casenames[classname]:
                left_indent = max(len(casename), left_indent)

        # render the axis headings in a stepped tree
        treelines = ""
        for filename in self.report_order():
            output += "{}    {}{}\n".format(" " * left_indent, treelines,
                                            filename)
            treelines += "| "
        output += "{}    {}\n".format(" " * left_indent, treelines)
        # render in groups of the same class

        for classname in self.classes:
            # new class
            output += "{}  \n".format(classname)

            # print the case name
            for casename in sorted(set(self.casenames[classname])):
                output += "- {}{}  ".format(casename,
                                            " " * (left_indent - len(casename)))

                # print each test and its result for each axis
                case_data = ""
                testcase: "Optional[Case]" = None
                for axis in self.report_order():
                    if axis not in self.cases[classname][casename]:
                        case_data += "  "
                    else:
                        testcase = self.cases[classname][casename][axis]
                        if testcase.skipped:
                            case_data += "s "
                        elif testcase.failure:
                            case_data += "f "
                        else:
                            case_data += "/ "

                if testcase is None or testcase.name is None:
                    testcase_name = ""
                else:
                    testcase_name = testcase.name
                combined, combined_name = self.combined_result(
                    self.case_results[classname][testcase_name])

                output += case_data
                output += " {} {}\n".format(combined, combined_name)

        # print the result stats

        output += "\n"
        output += "-" * 79
        output += "\n"

        output += "Test Results:\n"

        for outcome in sorted(self.result_stats):
            output += "  {:<12} : {:>6}\n".format(
                outcome.title(),
                self.result_stats[outcome])

        return output



"""
Classes for merging several reports into one
"""
from __future__ import unicode_literals

from typing import TYPE_CHECKING

import os
import xml.etree.ElementTree as ET
from io import BytesIO

from . import parser
from .render import ReportContainer
from .textutils import unicode_str

if TYPE_CHECKING: # pragma: no cover
    from typing import List


def is_junit_file(filepath: str):
    """
    Return True if this might be a junit xml file.
    :return:
    """
    try:
        with open(filepath, "r", encoding="utf-8") as fh:
            start = fh.read(512)
        for fragment in ["<testsuite", "<testrun"]:
            if fragment in start:
                return True
    except UnicodeDecodeError:
        pass
    return False



class Merger(ReportContainer, parser.ToJunitXmlBase):
    """
    Utility class to create a merged junix xml report
    """
    suites: "List[parser.Suite]"

    def __init__(self):
        super(Merger, self).__init__()
        self.suites = []

    def add_report(self, filename: str):
        """
        Load a test report or folder
        :param filename:
        :return:
        """
        if os.path.isfile(filename):
            report = parser.Junit(filename)
            self.reports[filename] = report
            for suite in report.suites:
                self.suites.append(suite)
        elif os.path.isdir(filename):
            # try importing all files in this folder
            for root, dirs, files in os.walk(filename):
                for filename in files:
                    filepath = os.path.join(root, filename)
                    try:
                        if is_junit_file(filepath):
                            try:
                                self.add_report(filepath)
                            except (parser.ParserError, ET.ParseError):
                                pass
                    except UnicodeDecodeError:
                        pass

    def add_suite(self, suite: "parser.Suite"):
        """
        Add a suite to the merge
        :param suite:
        :return:
        """
        self.suites.append(suite)

    def calculate_duration(self):
        """
        Add up the time values in all testcases
        :return:
        """
        total = 0
        for suite in self.suites:
            for testcase in suite.all():
                total += testcase.duration
        return total

    def tojunit(self):
        """
        Render a merged xml report
        :return:
        """
        root = self.make_element("testsuites")
        root.set(u"duration", unicode_str(self.calculate_duration()))
        for suite in self.suites:
            root.append(suite.tojunit())
        return root

    def toxmlstring(self):
        """
        Render the xml document as a string
        :return:
        """
        tree = ET.ElementTree(self.tojunit())
        buf = BytesIO()
        tree.write(buf)
        return u'<?xml version="1.0" encoding="utf-8"?>' + u"\n" + unicode_str(buf.getvalue())



"""
A series of tests that produce different example output
"""

import sys


def test_stderr_only():
    """
    Print some text to stderr
    :return:
    """
    sys.stderr.write("""
    Hello Standard Error
    =====================================

    This is some formatted stderr
    """)


def test_stdout_only():
    """
    Print some text to stderr
    :return:
    """
    sys.stdout.write("""
    Hello Standard Out
    =====================================

    This is some formatted stdout
    """)


def test_stdoe():
    """
    Print some stuff to stderr and stdout
    :return:
    """
    def err(msg):
        sys.stderr.write(msg)
        sys.stderr.write("\n")

    def out(msg):
        sys.stdout.write(msg)
        sys.stdout.write("\n")

    for _ in range(3):
        for word in ["Hello", "World"]:
            err("Err " + word)
            out("Out " + word)




"""
Test that we can load all our input files without error
"""
import pytest
from . import inputfiles
from junit2htmlreport import render, parserimpl


@pytest.mark.parametrize("filename", inputfiles.get_reports())
def test_load(filename):
    report = parserimpl.load_report(inputfiles.get_filepath(filename))
    assert report is not None

    doc = render.HTMLReport()
    doc.load(report, filename)

    output = str(doc)
    assert output



"""
Helper for loading our test files
"""
import os
HERE = os.path.dirname(os.path.abspath(__file__))


def get_reports():
    return [x for x in os.listdir(HERE) if x.endswith(".xml")]


def get_filepath(filename):
    """

    :param filename:
    :return:
    """
    return os.path.join(HERE, filename)



"""
Stringify to unicode
"""

from typing import Any, Optional


def unicode_str(text: "Optional[Any]"):
    """
    Convert text to unicode
    :param text:
    :return:
    """
    if isinstance(text, bytes):
        return text.decode("utf-8", "strict")
    return "" if text is None else str(text)



# -*- coding: utf-8 -*-
from junit2htmlreport import runner
from .inputfiles import get_filepath


def test_matrix_stdout(capsys):
    runner.run(["--summary-matrix", get_filepath("junit-unicode.xml")])

    out, err = capsys.readouterr()

    assert u"A Class with a cent ¢" in out
    assert u"Euro € Test Case" in out


from xml.etree import ElementTree

from junit2htmlreport import parser as j2h


def test_public_api():
    container = j2h.Junit(xmlstring="""<?xml version="1.0" encoding="UTF-8"?>
    <testsuite name="suite"></testsuite>
    """)
    container.filename = "test_results.xml"
    document = j2h.Suite()
    container.suites = [document]
    document.name = "test report"
    document.duration = 0.1
    document.package = "com.tests"
    first = j2h.Class()
    first.name = "myclass"
    document.classes[first.name] = first

    test1 = j2h.Case()
    test1.name = "test_one"
    test1.duration = 1.1
    test1.testclass = first
    first.cases.append(test1)

    test2 = j2h.Case()
    test2.name = "test_two"
    test2.duration = 1.2
    test2.testclass = first
    first.cases.append(test2)

    skipped1 = j2h.Case()
    skipped1.name = "test_skippy"
    skipped1.duration = 1.3
    skipped1.testclass = first
    skipped1.skipped = "test skipped"
    skipped1.skipped_msg = "test was skipped at runtime"
    first.cases.append(skipped1)

    failed1 = j2h.Case()
    failed1.name = "test_bad"
    failed1.duration = 1.4
    failed1.testclass = first
    failed1.failure = "test failed"
    failed1.failure_msg = "an exception happened"
    first.cases.append(failed1)

    html = container.html()

    assert html
    assert "<html" in html
    assert """test_skippy""" in html
    assert """an exception happened""" in html



"""
Tests for emitting junit xml
"""

from junit2htmlreport import parser, merge


def test_case_tojunit_failed():
    """
    Test coverting a failed case to xml
    :return:
    """
    testclass = parser.Class()
    testclass.name = "myclass"
    testcase = parser.Case()
    testcase.name = "mytest"
    testcase.testclass = testclass
    testcase.failure = "bad"
    testcase.failure_msg = "failed message"

    result = testcase.tojunit()
    children = list(result)
    assert len(children) == 1
    assert children[0].tag == "failure"
    assert children[0].text == "bad"
    assert children[0].get("message") == "failed message"


def test_case_tojunit():
    """
    Test converting a test case to a junit xml element
    :return:
    """
    testclass = parser.Class()
    testclass.name = "myclass"
    testcase = parser.Case()
    testcase.name = "mytest"
    testcase.testclass = testclass
    testcase.duration = 3.0
    testcase.stdout = "hello\nworld"
    testcase.stderr = "byee"
    testcase.skipped = "ignored"
    testcase.skipped_msg = "skipped message"

    result = testcase.tojunit()
    assert result.tag == "testcase"
    assert result.get("time") == "3.0"
    assert result.get("classname") == "myclass"
    assert result.get("name") == "mytest"
    children = list(result)
    assert len(children) == 3
    assert children[0].tag == "system-err"
    assert children[0].text == "byee"
    assert children[1].tag == "system-out"
    assert children[1].text == "hello\nworld"
    assert children[2].tag == "skipped"
    assert children[2].get("message") == "skipped message"
    assert children[2].text == "ignored"


def test_merge():
    """
    Test merging suites
    :return:
    """
    # make two suites, with two classes with each three cases
    suites = []
    for y in range(2):
        suite = parser.Suite()
        suites.append(suite)

        for x in range(2):
            testclass = parser.Class()
            testclass.name = "class.name{}_{}".format(y, x)
            suite.classes[testclass.name] = testclass
            for z in range(3):
                testcase = parser.Case()
                testcase.name = "testcase{}_{}_{}".format(y, x, z)
                testcase.testclass = testclass
                testclass.cases.append(testcase)
    merger = merge.Merger()
    for suite in suites:
        merger.add_suite(suite)

    result = merger.tojunit()
    assert result is not None

    output = merger.toxmlstring()
    assert output



"""Test the merge utility"""
import os
from junit2htmlreport import runner
from pathlib import Path
TESTS = Path(__file__).parent


def test_merge_folder(tmp_path):
    outfile = os.path.join(tmp_path, "test-merged.xml")
    outmatrix = os.path.join(tmp_path, "matrix.html")
    runner.run(["--merge", outfile, str(TESTS)])
    runner.run(["--theme", "dark", outfile])
    assert os.path.exists(f"{outfile}.html")

    tests = list([str(x) for x in TESTS.glob("*.xml")])
    runner.run(["--report-matrix", outmatrix] + tests)
    assert os.path.exists(outmatrix)



"""
Test the matrix functionality
"""

from junit2htmlreport import matrix
from junit2htmlreport.matrix import PARTIAL_PASS, PARTIAL_FAIL, TOTAL_FAIL, UNTESTED
from junit2htmlreport.parser import PASSED, SKIPPED, FAILED, UNKNOWN


from .inputfiles import get_filepath


def test_combined_result():
    """
    Test that the combined result string and short result value are correct
    :return:
    """
    textmatrix = matrix.TextReportMatrix()
    short, result = textmatrix.combined_result([PASSED, SKIPPED])

    assert short == textmatrix.short_outcome(PASSED)
    assert result == PASSED.title()

    short, result = textmatrix.combined_result([PASSED, FAILED])
    assert short == textmatrix.short_outcome(PARTIAL_FAIL)
    assert result == PARTIAL_FAIL.title()

    short, result = textmatrix.combined_result([PARTIAL_PASS])
    assert short == textmatrix.short_outcome(PARTIAL_PASS)
    assert result == PARTIAL_PASS.title()

    short, result = textmatrix.combined_result([TOTAL_FAIL])
    assert short == textmatrix.short_outcome(TOTAL_FAIL)
    assert result == TOTAL_FAIL.title()

    short, result = textmatrix.combined_result([FAILED, FAILED])
    assert short == textmatrix.short_outcome(FAILED)
    assert result == FAILED.title()

    short, result = textmatrix.combined_result([PASSED])
    assert short == textmatrix.short_outcome(PASSED)
    assert result == PASSED.title()

    short, result = textmatrix.combined_result([SKIPPED, SKIPPED])
    assert short == textmatrix.short_outcome(UNTESTED)
    assert result == UNTESTED.title()

    short, result = textmatrix.combined_result([])
    assert '?' == textmatrix.short_outcome(None) # type: ignore
    assert result == UNKNOWN.title()


def test_matrix_load(tmpdir):
    """
    Test loading multiple reports
    :return:
    """
    textmatrix = matrix.TextReportMatrix()
    textmatrix.add_report(get_filepath("junit-simple_suite.xml"))
    textmatrix.add_report(get_filepath("junit-simple_suites.xml"))
    textmatrix.add_report(get_filepath("junit-unicode.xml"))
    textmatrix.add_report(get_filepath("junit-unicode2.xml"))
    textmatrix.add_report(get_filepath("junit-cute2.xml"))
    textmatrix.add_report(get_filepath("junit-jenkins-stdout.xml"))

    assert len(textmatrix.reports) == 6

    result = textmatrix.summary()

    print(result)


def test_matrix_html(tmpdir):
    """
    Test loading multiple reports
    :return:
    """
    htmatrix = matrix.HtmlReportMatrix(str(tmpdir))
    htmatrix.add_report(get_filepath("junit-simple_suite.xml"))
    htmatrix.add_report(get_filepath("junit-simple_suites.xml"))
    htmatrix.add_report(get_filepath("junit-unicode.xml"))
    htmatrix.add_report(get_filepath("junit-axis-linux.xml"))
    assert len(htmatrix.reports) == 4
    result = htmatrix.summary()
    assert result.endswith("</html>")




"""
Render junit reports as HTML
"""
import os
import importlib.util
from typing import TYPE_CHECKING

if TYPE_CHECKING: # pragma: no cover
    from os import PathLike
    from typing import Optional, List, Dict

from jinja2 import Environment, PackageLoader, select_autoescape, FileSystemLoader, BaseLoader


class LoaderFactory:
    """Obtain jinja template loaders"""

    DEFAULT = "default"
    pkgname = "junit2htmlreport"

    THEME_DIR_ENV = "JUNIT2HTML_THEMES"
    THEME_FILES = ["base.html", "matrix.html", "report.html", "styles.css"]

    def __init__(self):
        self.baseloaders: "Dict[str, BaseLoader]" = {
            self.DEFAULT: PackageLoader(self.pkgname, os.path.join("templates", self.DEFAULT)),
        }

        spec = importlib.util.find_spec(self.pkgname)
        if spec and spec.origin:
            pkgdir = os.path.dirname(spec.origin)
            templates_dir = os.path.join(pkgdir, "templates")
            for name in os.listdir(templates_dir):
                if name not in self.baseloaders.keys():
                    self.baseloaders[name] = PackageLoader(self.pkgname, os.path.join("templates", name))

        extra_themes = os.getenv(self.THEME_DIR_ENV, None)
        if extra_themes and os.path.isdir(extra_themes):
            for root, folders, filenames in os.walk(os.path.abspath(extra_themes)):
                if filenames:
                    theme = True
                    for item in self.THEME_FILES:
                        if item not in filenames:
                            theme = False
                    if theme:
                        theme_name = os.path.basename(root)
                        self.baseloaders[theme_name] = FileSystemLoader(root)

        self.selected = self.DEFAULT

    def add_filesystem_loader(self, name: str, path: "PathLike"):
        self.baseloaders[name] = FileSystemLoader(path)

    def select_loader(self, name: str) -> None:
        if name in self.baseloaders.keys():
            self.selected = name

    def get_loader(self) -> BaseLoader:
        return self.baseloaders[self.selected]

    @property
    def styles(self) -> list:
        return sorted(list(self.baseloaders.keys()))


class JunitLoaderBase:
    """Contains a jinja template BaseLoader"""
    def __init__(self):
        self._loader_factory: "Optional[LoaderFactory]" = None

    @property
    def loader_factory(self) -> "LoaderFactory":
        if self._loader_factory is None:
            return LoaderFactory()
        return self._loader_factory

    @loader_factory.setter
    def loader_factory(self, value):
        self._loader_factory = value


class ReportContainer(JunitLoaderBase):
    """
    Hold one or more reports
    """
    reports: "Dict[str, Junit]"

    def __init__(self):
        super().__init__()
        self.reports = {}

    def add_report(self, filename: str) -> None:
        raise NotImplementedError()

    def failures(self):
        """
        Return all the failed test cases
        :return:
        """
        found: "List[Case]" = []
        for report in self.reports:
            for suite in self.reports[report].suites:
                found.extend(suite.failed())

        return found

    def skips(self):
        """
        Return all the skipped test cases
        :return:
        """
        found: "List[Case]" = []
        for report in self.reports:
            for suite in self.reports[report].suites:
                found.extend(suite.skipped())
        return found




class HTMLReport:

    def __init__(self, show_toc: bool=True):
        self.show_toc = show_toc
        self.title: str = ""
        self.report: "Optional[Junit]" = None

    def load(self, report: "Junit", title: str="JUnit2HTML Report"):
        self.report = report
        self.title = title

    def __iter__(self):
        if self.report is None:
            raise Exception("A report must be loaded through `load(...)` first.")

        return self.report.__iter__()

    def __str__(self) -> str:
        loader = self.report.loader_factory.get_loader()
        env = Environment(loader=loader, autoescape=select_autoescape(["html"]))
        template = env.get_template("report.html")
        return template.render(report=self, title=self.title, show_toc=self.show_toc)




from enum import Enum


class CaseResult(str, Enum):
    UNTESTED = "untested"
    PARTIAL_PASS = "partial pass"
    PARTIAL_FAIL = "partial failure"
    TOTAL_FAIL = "total failure"
    FAILED = "failed"  # the test failed
    SKIPPED = "skipped"  # the test was skipped
    PASSED = "passed"  # the test completed successfully
    ABSENT = "absent"  # the test was known but not run/failed/skipped
    UNKNOWN = ""

    def __str__(self) -> str:
        return self.value



"""
Test that does nothing other than import
"""
import os
import pytest
from .inputfiles import get_filepath
from .helpers import run_runner
from junit2htmlreport import parser, runner


def test_runner_sonobouy(tmpdir):
    """
    Test the stand-alone app with report produced by sonobouy
    :param tmpdir:  py.test tmpdir fixture
    :return:
    """
    run_runner(tmpdir, "junit-sonobouy.xml")


def test_runner_complex(tmpdir):
    """
    Test the stand-alone app with a large fairly complex junit file
    :param tmpdir:  py.test tmpdir fixture
    :return:
    """
    run_runner(tmpdir, "junit-complex_suites.xml")


def test_runner_6700(tmpdir):
    """
    Test the 6700 report
    I can't remember what is special about this file!
    :param tmpdir:
    :return:
    """
    run_runner(tmpdir, "junit-report-6700.xml")


def test_runner_unicode(tmpdir):
    """
    Test the stand-alone app with a unicode file (contains a euro symbol)
    :param tmpdir:
    :return:
    """
    run_runner(tmpdir, "junit-unicode.xml")


def test_runner_testrun(tmpdir):
    """
    Test the stand-alone app with a file rooted at <testrun>
    :param tmpdir:
    :return:
    """
    run_runner(tmpdir, "junit-testrun.xml")


def test_runner_merge(tmpdir):
    """
    Test merging multiple files
    :param tmpdir:
    :return:
    """
    filenames = ["junit-complex_suites.xml",
                 "junit-cute2.xml",
                 "junit-unicode.xml"]

    filepaths = []
    for filename in filenames:
        filepaths.append(
            os.path.join(tmpdir.strpath, get_filepath(filename)))
    newfile = os.path.join(tmpdir.strpath, "merged.xml")
    args = ["--merge", newfile]
    args.extend(filepaths)
    runner.run(args)
    assert os.path.exists(newfile)


def test_emit_stdio():
    """
    Test the stand-alone app can generate a page from a report containing stdio text
    But also save the result in the current folder
    :return:
    """
    folder = os.path.dirname(__file__)
    reportfile = os.path.join(folder, "junit-jenkins-stdout.xml")
    runner.run([reportfile])
    htmlfile = os.path.join(folder, "junit-jenkins-stdout.xml.html")
    assert os.path.exists(htmlfile)
    with open(htmlfile, "r") as readfile:
        content = readfile.read()
        assert "===&gt; Executing test case" in content


def test_parser():
    """
    Test the junit parser directly
    :return:
    """
    junit = parser.Junit(filename=get_filepath("junit-simple_suite.xml"))
    assert len(junit.suites) == 1
    assert len(junit.suites[0].properties) == 3

    junit = parser.Junit(filename=get_filepath("junit-simple_suites.xml"))
    assert len(junit.suites) == 1
    assert len(junit.suites[0].properties) == 3

    junit = parser.Junit(filename=get_filepath("junit-complex_suites.xml"))
    assert len(junit.suites) == 66

    junit = parser.Junit(filename=get_filepath("junit-cute2.xml"))
    assert len(junit.suites) == 6

    junit = parser.Junit(filename=get_filepath("junit-unicode.xml"))
    assert len(junit.suites) == 1
    assert len(junit.suites[0].classes) == 2

    # different report structure, both files contain unicode symbols
    junit = parser.Junit(filename=get_filepath("junit-unicode2.xml"))
    assert len(junit.suites) == 1
    assert len(junit.suites[0].classes) == 1


def test_binary_names():
    # a test with nonsense binary in the case names
    junit = parser.Junit(filename=get_filepath("pytest-binary-names.xml"))
    assert junit


def test_parser_stringreader():
    """
    Test the junit parser when reading strings
    :return:
    """
    with open(get_filepath("junit-complex_suites.xml"), "r") as data:
        junit = parser.Junit(xmlstring=data.read())
        assert len(junit.suites) == 66
        assert junit.suites[0].name == "Untitled suite in /Users/niko/Sites/casperjs/tests/suites/casper/agent.js"
        assert junit.suites[0].package == "tests/suites/casper/agent"
        assert junit.suites[0].classes["tests/suites/casper/agent"].cases[1].name == "Default user agent matches /plop/"


def test_fail_exit(tmpdir):
    """
    Test the tool exits with an error for --max-failures
    :return:
    """
    with pytest.raises(SystemExit) as err:
        run_runner(tmpdir, "junit-unicode.xml", "--summary-matrix", "--max-failures", "1")

    assert err.value.code != 0


def test_skip_exit(tmpdir):
    """
    Test the tool exits with an error for --max-skipped
    :return:
    """
    with pytest.raises(SystemExit) as err:
        run_runner(tmpdir, "junit-axis-windows.xml", "--summary-matrix", "--max-skipped", "1")

    assert err.value.code != 0


"""
Helper funcs for tests
"""
import os
from .inputfiles import get_filepath
from junit2htmlreport import runner


def run_runner(tmpdir, filename, *extra):
    """
    Run the junit2html program against the given report and produce a html doc
    :param tmpdir:
    :param filename:
    :param extra: addtional arguments
    :return:
    """
    testfile = get_filepath(filename=filename)
    if not len(extra):
        outfile = os.path.join(tmpdir.strpath, "report.html")
        runner.run([testfile, outfile])
        assert os.path.exists(outfile)
    else:
        runner.run([testfile] + list(extra))


def test_runner_simple(tmpdir):
    """
    Test the stand-alone app with a simple fairly empty junit file
    :param tmpdir:  py.test tmpdir fixture
    :return:
    """
    run_runner(tmpdir, "junit-simple_suites.xml")



"""
Small command line tool to generate a html version of a junit report file
"""
from typing import TYPE_CHECKING

import os
import sys
from argparse import ArgumentParser

from . import matrix, merge, parser, render

if TYPE_CHECKING:
    from typing import List

loader_factory = render.LoaderFactory()
installed_themes = sorted(list(loader_factory.styles))

PARSER = ArgumentParser(prog="junit2html")

PARSER.add_argument("--summary-matrix", dest="text_matrix", action="store_true",
                    default=False,
                    help="Render multiple result files to the console")

PARSER.add_argument("--report-matrix", dest="html_matrix", type=str,
                    metavar="REPORT",
                    help="Generate an HTML report matrix")

PARSER.add_argument("--max-failures", dest="fail", type=int, default=0,
                    metavar="FAILURES",
                    help="Exit non-zero if FAILURES or more test cases are failures (has no effect with --merge)")

PARSER.add_argument("--max-skipped", dest="skip", type=int, default=0,
                    metavar="SKIPPED",
                    help="Exit non-zero if SKIPPED or more test cases are skipped (has no effect with --merged)")

PARSER.add_argument("--merge", dest="merge_output", type=str,
                    metavar="NEWREPORT",
                    help="Merge multiple test results into one file")
THEME_MUTEX = PARSER.add_mutually_exclusive_group()
THEME_MUTEX.add_argument("--theme", type=str, dest="theme", choices=installed_themes,
                    help="Use an installed junit2html theme")

THEME_MUTEX.add_argument("--theme-dir", type=str, dest="template_folder",
                    help="Use a folder containing custom html and CSS.")

PARSER.add_argument("--reports-template-folder", dest="template_folder", type=str,
                    help="Render reports with these templates (deprecated, use '--theme-dir' instead")

PARSER.add_argument("--hide-toc", dest="hide_toc", action="store_true",
                    default=False,
                    help="Don't include a table-of-contents in the HTML report")

PARSER.add_argument("REPORTS", metavar="REPORT", type=str, nargs="+",
                    help="Test file to read")

PARSER.add_argument("OUTPUT", type=str, nargs="?",
                    help="Filename to save the html as")


def run(args: "List[str]"):
    """
    Run this tool
    :param args:
    :return:
    """
    opts = PARSER.parse_args(args) if args else PARSER.parse_args()
    if opts.template_folder and os.path.isdir(opts.template_folder):
        loader_factory.add_filesystem_loader(loader_factory.DEFAULT, os.path.abspath(opts.template_folder))
    if opts.theme:
        loader_factory.select_loader(opts.theme)
    inputs = opts.REPORTS
    util = None
    if opts.merge_output:
        util = merge.Merger()
        for inputfile in inputs:
            util.add_report(inputfile)

        xmltext = util.toxmlstring()
        with open(opts.merge_output, "w") as outfile:
            outfile.write(xmltext)
    elif opts.text_matrix:
        util = matrix.TextReportMatrix()
        for filename in inputs:
            util.add_report(filename)
        print(util.summary())
    elif opts.html_matrix:
        util = matrix.HtmlReportMatrix(os.path.dirname(opts.html_matrix))
        util.loader_factory = loader_factory
        for filename in inputs:
            util.add_report(filename, show_toc=not opts.hide_toc)
        with open(opts.html_matrix, "w") as outfile:
            outfile.write(util.summary())

    if util:
        if opts.fail:
            failed = util.failures()
            if len(failed) >= opts.fail:
                sys.exit(len(failed))
        if opts.skip:
            skipped = util.skips()
            if len(skipped) >= opts.fail:
                sys.exit(len(skipped))

    if not util:
        # legacy interface that we need to preserve
        # no options, one or two args, first is input file, optional second is output

        if len(opts.REPORTS) > 2:
            PARSER.print_usage()
            sys.exit(1)

        infilename = opts.REPORTS[0]

        if len(opts.REPORTS) == 2:
            outfilename = opts.REPORTS[1]
        else:
            outfilename = infilename + ".html"

        report = parser.Junit(infilename)
        report.loader_factory = loader_factory
        html = report.html(show_toc=not opts.hide_toc)
        if report.filename is not None:
            with open(outfilename, "wb") as outfile:
                outfile.write(html.encode('utf-8'))
        else:
            print(html.encode('utf-8'))


def start():
    """
    Run using the current sys.argv
    """
    run(sys.argv[1:])


if __name__ == "__main__":
    start()
